     /**
      * \brief Property files
      *
      * The Properties service program provides procedures to work on/with a 
      * persistent set of properties (key/value pair). Properties can be loaded 
      * from or saved to a stream file. Keys and Values are both of data type 
      * string. Each key can only exist once in a set of properties.
      * <br><br>
      * The <em>load</em> procedure will load all properties from a stream file.
      * LF (x'25') and CRLF (x'0D25') are supported as line separation characters.
      * All natively supported codepages can be processed by the load procedure.
      * 
      * \author Mihael Schmidt
      * \date   29.04.2011
      *
      * \todo save properties to stream file  
      */

      *---------------------------------------------------------------------------------------------
      *
      * (C) Copyleft 2011 Mihael Schmidt
      *
      * This file is part of PROPERTIES project and service program.
      *
      * PROPERTIES is free software: you can redistribute it and/or modify
      * it under the terms of the GNU Lesser General Public License as published
      * by the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * PROPERTIES is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU Lesser General Public License for more details.
      *
      * You should have received a copy of the GNU Lesser General Public License
      * along with PROPERTIES.  If not, see <http://www.gnu.org/licenses/>.
      *
      *---------------------------------------------------------------------------------------------


     H nomain

      *-------------------------------------------------------------------------
      * Prototypes
      *-------------------------------------------------------------------------
      /include 'properties_h.rpgle'
      /include 'rng/input_provider_h.rpgle'
      /include 'rng/streamfile_input_provider_h.rpgle'
      /include 'libtree/libtree_h.rpgle'
      /include 'arraylist_h.rpgle'
      /include 'libc_h.rpgle'
 
 
      *-------------------------------------------------------------------------
      * Procedures
      *-------------------------------------------------------------------------
 
     /**
      * \brief Create new properties object
      *
      * \return Pointer to new properties object
      */
     P properties_create...
     P                 B                   export
     D                 PI              *
      *
     D properties      S               *
      /free
       return tree_rb_create(%paddr('tree_rb_string_compare'));
      /end-free
     P                 E
 
 
     /**
      * \brief Load property file
      *
      * Loads the properties from the passed file or alternative input provider.
      *
      * \author Mihael Schmidt
      * \date   29.04.2011
      *
      * \param Path to proptery file (optional if alt. input provider passed)
      * \param Separator (default: = )
      * \param Alternative Input Provider
      *
      * \return Pointer to properties object or *null if an error occured
      */
     P properties_load...
     P                 B                   export
     D                 PI              *
     D   pPath                     1024A   const varying options(*omit)
     D   pSeparator                   1A   const options(*nopass : *omit)
     D   pInput                        *   const options(*nopass)
      *
     D properties      S               *
     D path            S           1024A
     D separator       S              1A
      *
     D input           S               *
     D fd              S             10I 0
     D newline         S              1A   inz(x'25')
     D line            S          65535A   varying
     D lineLength      S             10I 0
     D value           S          65535A   varying
     D x               S             10I 0
     D key             S          65535A   varying
     D extended        S               N   inz(*off)
     D contLine        S               N
      /free
       if (%addr(pPath) <> *null);
         path = pPath + x'00';
       endif;
 
       if (%parms() >= 2 and %addr(pSeparator) <> *null);
         separator = pSeparator;
       else;
         separator = '=';
       endif;
 
       if (%parms() = 3);
         input = pInput;
       else;
         input = rng_input_stmf_create(%addr(path));
       endif;
 
       properties = properties_create();
 
       // datei ffnen
       monitor;
         rng_input_stmf_open(input);
 
         // read text from property file by line
         dow ( rng_input_read(input : line));
           if (line = *blank);
             // ignore blank lines
 
           elseif (%subst(%triml(line) : 1 : 1) = '!' or
                   %subst(%triml(line) : 1 : 1) = '#');
             // ignore comments (! and #)
                          
           else;
 
             // check for separator
             x = %scan(separator : line);
             if (x > 0 or extended);
               // keep the "old" key if there was a continuation
               if (not extended);
                 key = %trim(%subst(line : 1 : x - 1));
               endif;
 
               lineLength = %len(%trimr(line));
 
               // check if last character is a continuation character (backslash)
               if (%subst(line : lineLength : 1) = '\');
                 contLine = *on;
                 extended = *on;
                 value += %trim(%subst(line : x + 1));
                 %len(value) = %len(value) -1;
               else;
                 value += %trim(%subst(line : x + 1));
               endif;
 
               if (not contLine);
                 tree_rb_string_put(properties : key :
                                    %addr(value : *DATA) : %len(value));
                 %len(value) = 0;
                 extended = *off;
               endif;
 
             else;
               // only the key with no value => add it this way
               key = %trim(line);
               tree_rb_string_put(properties : key :
                                  %addr(value : *DATA) : %len(value));
             endif;
           endif;
 
           // clear line
           line = *blank;
           contLine = *off;
         enddo;
 
         on-error *all;
           // return an empty properties object
           // TODO send specific escape message
           properties_finalize(properties);
           properties = *null;
       endmon;
 
       rng_input_close(input);
       rng_input_finalize(input);
 
       return properties;
      /end-free
     P                 E
 
 
     /**
      * \brief Dispose properties object
      *
      * Releases all allocated resources.
      *
      * \param Pointer to properties object
      */
     P properties_finalize...
     P                 B                   export
     D                 PI
     D   properties                    *
      /free
       tree_rb_string_finalize(properties);
      /end-free
     P                 E
 
 
     /**
      * \brief Get property value
      *
      * Returns the value to the passed property key.
      * <br><br>
      * If the property does not exist and a default value is passed
      * the default value will be returned.
      *
      * \param Pointer to properties object
      * \param Key
      * \param Default value (optional)
      *
      * \return Property value or *blank if the property does not exist.
      *
      * \todo change to escape message if property doesn't exist
      */
     P properties_get...
     P                 B                   export
     D                 PI         65535A   varying
     D   properties                    *   const
     D   key                      65535A   const varying
     D   default                  65535A   const varying options(*nopass)
      *
     D node            DS                  likeds(tree_node_string_t) based(ptr)
      /free
       ptr = tree_rb_string_get(properties : key);
       if (ptr <> *null);
         return %str(node.value);
       elseif (%parms() = 3);
         return default;
       else;
         // TODO send escape message
         return *blank; // dummy : will never come here
       endif;
      /end-free
     P                 E
 
 
     /**
      * \brief Add property
      *
      * Adds the passed property to the properties object. If a property with
      * the passed key already exists the value of the property will be
      * replaced with the passed value.
      *
      * \param Pointer to properties object
      * \param Key
      * \param Value
      */
     P properties_put...
     P                 B                   export
     D                 PI
     D   properties                    *   const
     D   key                      65535A   const varying
     D   pValue                   65535A   const varying
      *
     D value           S          65535A   varying
      /free
       value = pValue;
       tree_rb_string_put(properties : key : %addr(value:*DATA) : %len(value));
      /end-free
     P                 E
 
 
     /**
      * \brief Clear properties object
      *
      * Removes all properties from the object.
      *
      * \param Pointer to properties object
      */
     P properties_clear...
     P                 B                   export
     D                 PI
     D   properties                    *   const
      *
      /free
       tree_rb_string_clear(properties);
      /end-free
     P                 E
 
 
     /**
      * \brief Check property existence
      *
      * Checks if the properties object contains the passed property key.
      *
      * \param Pointer to properties object
      * \param Key
      *
      * \return *on = property exists otherwise false
      */
     P properties_contains...
     P                 B                   export
     D                 PI              N
     D   properties                    *   const
     D   key                      65535A   const varying
      /free
       return tree_rb_string_containsKey(properties : key);
      /end-free
     P                 E
 
 
     /**
      * \brief Dump properties
      *
      * Outputs all properties to the screen (via opcode DSPLY).
      *
      * \param Pointer to properties object
      */
     P properties_dump...
     P                 B                   export
     D                 PI
     D  properties                     *   const
      *
     D displayedKeyLength...
     D                 S             10I 0 
     D dsp             S             50A
     D node            DS                  likeds(tree_node_string_t) based(ptr)
      /free
       ptr = tree_rb_string_first(properties);
       dow (ptr <> *null);
         clear dsp;
         if (node.keyLength > 50);
           displayedKeyLength = 50;
         else;
           displayedKeyLength = node.keyLength;
         endif;
         
         memcpy(%addr(dsp) : node.key : displayedKeyLength);
         
         dsp = %trimr(dsp) + ' - ' + %str(node.value);
         dsply dsp;
         ptr = tree_rb_string_next(ptr);
       enddo;
      /end-free
     P                 E
 
 
     /**
      * \brief List property keys
      *
      * Returns a list with the property keys from this properties object.
      *
      * \param Pointer to properties object
      *
      * \return List with property keys
      *
      * \info The resources of the returned list must be released by the caller
      *       with the procedure <em>arraylist_finalize</em>.
      */
     P properties_list...
     P                 B                   export
     D                 PI              *
     D  properties                     *   const
      *
     D keys            S               *
     D node            DS                  likeds(tree_node_string_t) based(ptr)
      /free
       keys = arraylist_create();
 
       ptr = tree_rb_string_first(properties);
       dow (ptr <> *null);
         arraylist_add(keys : node.key : node.keyLength);
         ptr = tree_rb_string_next(ptr);
       enddo;
 
       return keys;
      /end-free
     P                 E
 
 
     /**
      * \brief Get number of properties
      *
      * Returns the number of properties in this properties object.
      *
      * \param Pointer to properties object
      *
      * \return Number of properties
      */
     P properties_size...
     P                 B                   export
     D                 PI            10I 0
     D  properties                     *   const
      /free
       return tree_rb_size(properties);
      /end-free
     P                 E
 